#region Using declarations

// ReSharper disable RedundantUsingDirective
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Xml.Serialization;
using NinjaTrader.Cbi;
using NinjaTrader.Gui;
using NinjaTrader.Gui.Chart;
using NinjaTrader.Gui.SuperDom;
using NinjaTrader.Gui.Tools;
using NinjaTrader.Data;
using NinjaTrader.NinjaScript;
using NinjaTrader.Core.FloatingPoint;
using NinjaTrader.Gui.Tools;
using NinjaTrader.NinjaScript.Indicators;
using NinjaTrader.NinjaScript.Indicators.ABPowerX;

// ReSharper restore RedundantUsingDirective

#endregion

#region LICENSE

/*
"AB-PowerX" and associated add-on file is a NinjaTrader Indicator designed to provide a paint-bar indication of trend direction and also provide trade signal detection and automatic risk/reward display. This indicator is based on Markus Heitkoetter's PowerX System.

Copyright 2021 Asif Bacchus <asif@asifbacchus.dev>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

This software was designed and authored by Asif Bacchus and has also been publicly released under the terms as stated above ('MIT License'). Please contact the author directly at asif@asifbacchus.dev for further information or to inquire about proprietary licensing. Please refer to the private git repo at 'https://git.asifbacchus.dev/asif/NinjaTrader-PowerX.git' to submit enhancements, suggestions, bug reports, etc.
*/

#endregion

//This namespace holds Add ons in this folder and is required. Do not change it. 
// ReSharper disable once CheckNamespace
namespace NinjaTrader.NinjaScript.AddOns
{
    // ReSharper disable once InconsistentNaming
    // ReSharper disable once RedundantNameQualifier
    public class ABPowerX : NinjaTrader.NinjaScript.AddOnBase
    {
        private const string MyName = @"ABPowerXTools";
        private const string MyDescription = @"Addon classes and methods for AB-PowerX indicator.";

        protected override void OnStateChange()
        {
            // ReSharper disable once SwitchStatementMissingSomeEnumCasesNoDefault
            switch (State)
            {
                case State.SetDefaults:
                    Description = MyDescription;
                    Name = MyName;
                    break;
                case State.Configure:
                    break;
            }
        }
    }

    namespace AB_PowerX
    {
        #region Classes

        /// <summary>
        /// Object that stores information about PowerX Bars.
        /// </summary>
        internal class PowerXBar
        {
            private readonly Bars bars;
            private readonly double tickSize;
            internal int Index { get; private set; }
            internal int Expires { get; private set; }
            internal BarTrend Directionality { get; private set; }

            internal double ActivationValue { get; private set; }
            private bool Triggered { get; set; }
            internal double StopValue { get; private set; }
            internal double TargetValue { get; private set; }
            internal int RiskTicks { get; private set; }
            internal double RiskDollars { get; private set; }
            internal bool Active { get; private set; }


            /// <summary>
            /// Constructor for PowerXBars: Creates a PowerXBar object with the information it needs to process all its states.
            /// </summary>
            /// <param name="barsObj">Bar data from indicator class.</param>
            /// <param name="tickSz">Tick-size of instrument linked to chart where indicator is located.</param>
            /// <param name="triggerDistance">Ticks added/subtracted from Activation point the exceeding of which causes Activation.</param>
            /// <param name="currentBar">Index of current bar.</param>
            /// <param name="validityPeriod">Number of bars for which this object may be Triggered without becoming Active.</param>
            /// <param name="macd">MACD indicator used in directionality determination.</param>
            /// <param name="stochastics">Stochastics indicator used in directionality determination.</param>
            /// <param name="rsi">RSI indicator used in directionality determination.</param>
            public PowerXBar(Bars barsObj, double tickSz, int triggerDistance, int currentBar, int validityPeriod,
                MACD macd, Stochastics stochastics, RSI rsi)
            {
                bars = barsObj;
                tickSize = tickSz;
                Index = currentBar;
                Expires = currentBar + validityPeriod;
                Directionality = GetDirectionality(macd, stochastics, rsi);
                Triggered = false;
                ActivationValue = ((Func<double>)(() =>
                {
                    // ReSharper disable once SwitchStatementHandlesSomeKnownEnumValuesWithDefault
                    switch (Directionality)
                    {
                        case BarTrend.Bullish:
                            return bars.GetHigh(Index) + triggerDistance * tickSize;
                        case BarTrend.Bearish:
                            return bars.GetLow(Index) - triggerDistance * tickSize;
                        default:
                            return 0;
                    }
                }))();
                Active = false;
            }

            /// <summary>
            /// Set PowerXBar as 'triggered' and make risk and target information available.
            /// </summary>
            /// <param name="adrValue">Average Daily (bar) Range value.</param>
            /// <param name="stopMultiple">ADR multiplier value to calculate stop level.</param>
            /// <param name="targetMultiple">ADR multiplier value to calculate target level.</param>
            internal void Trigger(double adrValue, double stopMultiple, double targetMultiple)
            {
                Triggered = true;

                // calculate stop and target prices
                StopValue = Directionality == BarTrend.Bullish
                    ? bars.Instrument.MasterInstrument.RoundToTickSize(ActivationValue - stopMultiple * adrValue)
                    : bars.Instrument.MasterInstrument.RoundToTickSize(ActivationValue + stopMultiple * adrValue);
                TargetValue = Directionality == BarTrend.Bullish
                    ? bars.Instrument.MasterInstrument.RoundToTickSize(ActivationValue + targetMultiple * adrValue)
                    : bars.Instrument.MasterInstrument.RoundToTickSize(ActivationValue - targetMultiple * adrValue);

                // calculate risk
                var riskPoints =
                    bars.Instrument.MasterInstrument.RoundToTickSize(Math.Abs(ActivationValue - StopValue));
                var tickValue = bars.Instrument.MasterInstrument.PointValue * tickSize;
                RiskTicks = Convert.ToInt32(riskPoints / tickSize);
                RiskDollars = RiskTicks * tickValue;
            }

            /// <summary>
            /// Mark this PowerXBar as 'Active', meaning that its ActivationPrice has been exceeded.
            /// </summary>
            internal void Activate()
            {
                Active = true;
            }

            /// <summary>
            /// Return trend direction based on comparison on supplied indicators' values against fixed threshold values.
            /// </summary>
            /// <param name="macd">MACD indicator object.</param>
            /// <param name="stochastics">Stochastics object.</param>
            /// <param name="rsi">RSI object.</param>
            /// <returns>enum:BarTrend value.</returns>
            private static BarTrend GetDirectionality(MACD macd, Stochastics stochastics, RSI rsi)
            {
                var macdState = DoComparison(macd.Diff[0], 0d);
                var stochasticState = DoComparison(stochastics.K[0], 50d);
                var rsiState = DoComparison(rsi[0], 50d);

                switch (macdState + stochasticState + rsiState)
                {
                    case 3:
                        return BarTrend.Bullish;
                    case -3:
                        return BarTrend.Bearish;
                    default:
                        return BarTrend.Neutral;
                }
            }

            /// <summary>
            /// Compares input value against supplied threshold value and returns numerical representation of result.
            /// </summary>
            /// <param name="value">Value to evaluate against threshold.</param>
            /// <param name="threshold">Level to which 'value' is compared.</param>
            /// <returns>Int: -1 = value below threshold, 0 = value equals threshold, 1 = value above threshold.</returns>
            private static int DoComparison(double value, double threshold)
            {
                if (value > threshold) return 1;
                if (value < threshold) return -1;
                return 0;
            }
        }

        #endregion

        #region Enumerations

        /// <summary>
        /// Trend-types any particular bar may reflect at a given time.
        /// </summary>
        public enum BarTrend
        {
            Bearish,
            Bullish,
            Neutral
        }

        /// <summary>
        /// Directionality options for determining when to "fill" a candlestick.
        /// </summary>
        public enum CandleFillStyle
        {
            [Description(@"Close < Open")] FillDown,
            [Description(@"Close > Open")] FillUp
        }

        /// <summary>
        /// Valid label locations relative to the entity they are labelling.
        /// </summary>
        public enum TradePlannerLabelLocations
        {
            None,
            Default,
            Inverted
        }

        /// <summary>
        /// Valid options for information a label will convey.
        /// </summary>
        public enum TradePlannerLabelOptions
        {
            [Description(@"Qty | Price")] QtyPriceSplit,
            [Description(@"<none> | Qty @ Price")] QtyPrice,

            [Description(@"Risk in ticks | Qty @ Price")]
            RiskTicks,

            [Description(@"Risk in dollars | Qty @ Price")]
            RiskDollar,

            [Description(@"RiskT Risk$ | Qty @ Price")]
            All
        }

        /// <summary>
        /// Valid options for which trigger-lines should be displayed.
        /// </summary>
        public enum TriggerDisplayOptions
        {
            [Description(@"Show all trigger lines")]
            ShowAll,

            [Description(@"Remove expired trigger lines")]
            RemoveExpired,

            [Description(@"Only most recent trigger line")]
            MostRecent
        }

        #endregion

        #region TypeConverters

        /// <summary>
        /// Convert Enumeration Description decorations to NinjaTrader-readable descriptions for drop-down box text.
        /// reference:
        /// https://ninjatrader.com/support/forum/forum/ninjatrader-8/indicator-development/93865-how-do-i-implement-decorated-enums
        /// </summary>
        public class EnumDescriptionTypeConverter : EnumConverter
        {
            private readonly Type enumType;

            public EnumDescriptionTypeConverter(Type type) : base(type)
            {
                enumType = type;
            }

            public override bool CanConvertFrom(ITypeDescriptorContext context, Type srcType)
            {
                return srcType == typeof(string);
            }

            public override bool CanConvertTo(ITypeDescriptorContext context, Type destType)
            {
                return destType == typeof(string);
            }

            public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture,
                object value)
            {
                foreach (var fi in enumType.GetFields())
                {
                    var dna =
                        (DescriptionAttribute)Attribute.GetCustomAttribute(fi, typeof(DescriptionAttribute));
                    if (dna != null && (string)value == dna.Description)
                        return Enum.Parse(enumType, fi.Name);
                }

                // ReSharper disable once AssignNullToNotNullAttribute
                return Enum.Parse(enumType, (string)value);
            }

            public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture,
                object value, Type destType)
            {
                // ReSharper disable once PossibleNullReferenceException
                var fi = enumType.GetField(value.ToString());
                var dna =
                    (DescriptionAttribute)Attribute.GetCustomAttribute(fi, typeof(DescriptionAttribute));
                return dna != null ? dna.Description : value.ToString();
            }

            public override StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
            {
                var values = (from object v in Enum.GetValues(enumType)
                    let fi = enumType.GetField(v.ToString())
                    let dna = (DescriptionAttribute)Attribute.GetCustomAttribute(fi, typeof(DescriptionAttribute))
                    select dna != null ? dna.Description : v.ToString()).ToList();

                return new StandardValuesCollection(values);
            }
        }

        #endregion
    }
}